/*
Quake GameCube port.
Copyright (C) 2007 Peter Mackay
Copyright (C) 2008 Eluan Miranda

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// ELUTODO: do something about lookspring and lookstrafe
// ELUTODO: keys to: nunchuk turn and nunchuk look up/down?
// ELUTODO: osk doesn't work if client disconnected

#include "../generic/quakedef.h"

// pass these values to whatever subsystem wants it
float in_pitchangle;
float in_yawangle;
float in_rollangle;

// Are we inside the on-screen keyboard? (ELUTODO: refactor)
int in_osk = 0;

// \0 means not mapped...
// 5 * 15
char osk_normal[75] =
{
	'\'', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '=', ']', K_BACKSPACE,
	0, 'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', 0, '[', K_ENTER, K_ENTER,
	0, 'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', 0, '~', '/', K_ENTER, K_ENTER,
	0, '\\', 'z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', ';', K_ENTER, K_ENTER, K_ENTER,
	0 , 0, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, 0, 0
};

char osk_shifted[75] =
{
	'\"', '!', '@', '#', '$', '%', 0, '&', '*', '(', ')', '_', '+', '}', K_BACKSPACE,
	0, 'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', 0, '{', K_ENTER, K_ENTER,
	0, 'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 0, '^', '?', K_ENTER, K_ENTER,
	0, '|', 'Z', 'X', 'C', 'V', 'B', 'N', 'M', '<', '>', ':', K_ENTER, K_ENTER, K_ENTER,
	0 , 0, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, K_SPACE, 0, 0
};

char *osk_set;
int osk_selected;
int osk_last_selected;
int osk_coords[2];
float osk_repeat_delay = 1.0f;
float osk_last_press_time = 0.0f;

#include <ogc/pad.h>
#include <wiiuse/wpad.h>
#include "input_wiimote.h"

#define FORCE_KEY_BINDINGS 0

u32 wiimote_ir_res_x;
u32 wiimote_ir_res_y;

// wiimote info
u32 sys_previous_keys;
u32 sys_keys;

ir_t pointer;
orient_t orientation;
expansion_t expansion;

bool wiimote_connected = true;
bool nunchuk_connected = true;

int last_irx = -1, last_iry = -1;

static float clamp(float value, float minimum, float maximum)
{
	if (value > maximum)
	{
		return maximum;
	}
	else if (value < minimum)
	{
		return minimum;
	}
	else
	{
		return value;
	}
}

static void apply_dead_zone(float* x, float* y, float dead_zone)
{
	// Either stick out of the dead zone?
	if ((fabsf(*x) >= dead_zone) || (fabsf(*y) >= dead_zone))
	{
		// Nothing to do.
	}
	else
	{
		// Clamp to the dead zone.
		*x = 0.0f;
		*y = 0.0f;
	}
}

/*
 * [QUOTE=eke-eke]
 * the nunchuk analog stick seems to act weird sometime, here's the functions I added for reading the Sticks values like the old PAD_ library, maybe you would find this useful:
 */
static s8 WPAD_StickX(int wpad_exp, u8 which)
{
	float mag = 0.0;
	float ang = 0.0;

	switch (wpad_exp)
	{
		case WPAD_EXP_NUNCHUK:
		case WPAD_EXP_GUITARHERO3:
			if (which == 0)
			{
				mag = expansion.nunchuk.js.mag;
				ang = expansion.nunchuk.js.ang;
			}
			break;

		case WPAD_EXP_CLASSIC:
			if (which == 0)
			{
				mag = expansion.classic.ljs.mag;
				ang = expansion.classic.ljs.ang;
			}
			else
			{
				mag = expansion.classic.rjs.mag;
				ang = expansion.classic.rjs.ang;
			}
			break;

		default:
			break;
	}

	/* calculate X value (angle needs to be converted into radians) */
	if (mag > 1.0) mag = 1.0;
	else if (mag < -1.0) mag = -1.0;
	double val = mag * sin(M_PI * ang/180.0f);

	return (s8)(val * 128.0f);
}

static s8 WPAD_StickY(int wpad_exp, u8 which)
{
	float mag = 0.0;
	float ang = 0.0;

	switch (wpad_exp)
	{
		case WPAD_EXP_NUNCHUK:
		case WPAD_EXP_GUITARHERO3:
			if (which == 0)
			{
				mag = expansion.nunchuk.js.mag;
				ang = expansion.nunchuk.js.ang;
			}
			break;

		case WPAD_EXP_CLASSIC:
			if (which == 0)
			{
				mag = expansion.classic.ljs.mag;
				ang = expansion.classic.ljs.ang;
			}
			else
			{
				mag = expansion.classic.rjs.mag;
				ang = expansion.classic.rjs.ang;
			}
			break;

		default:
			break;
	}

	/* calculate X value (angle need to be converted into radian) */
	if (mag > 1.0) mag = 1.0;
	else if (mag < -1.0) mag = -1.0;
	double val = mag * cos(M_PI * ang/180.0f);

	return (s8)(val * 128.0f);
}


void IN_Init (void)
{
#if FORCE_KEY_BINDINGS
	// Set up the key bindings.
	Cbuf_AddText("bind JOY1 +jump\n");
	Cbuf_AddText("bind JOY2 +speed\n");
	Cbuf_AddText("bind JOY4 \"impulse 10\"\n");
	Cbuf_AddText("bind JOY6 +attack\n");
#endif

	last_irx = -1;
	last_iry = -1;

	in_osk = 0;

	in_pitchangle = .0f;
	in_yawangle = .0f;
	in_rollangle = .0f;

	WPAD_SetDataFormat(WPAD_CHAN_ALL, WPAD_FMT_BTNS_ACC_IR);
	WPAD_SetVRes(WPAD_CHAN_ALL, wiimote_ir_res_x, wiimote_ir_res_y);
}

void IN_Shutdown (void)
{
}

void IN_Commands (void)
{
	// Fetch the pad state.
	WPAD_ScanPads();
	sys_keys = WPAD_ButtonsHeld(WPAD_CHAN_0);
	WPAD_IR(WPAD_CHAN_0, &pointer);
	WPAD_Orientation(WPAD_CHAN_0, &orientation);
	WPAD_Expansion(WPAD_CHAN_0, &expansion);

	wiimote_connected = true;
	nunchuk_connected = true;

	if (wiimote_connected && (sys_keys & WPAD_BUTTON_MINUS))
	{
		// ELUTODO: we are using the previous frame wiimote position... FIX IT
		in_osk = 1;
		int line = (last_iry - OSK_YSTART) / (osk_line_size * (osk_line_size / osk_charsize)) - 1;
		int col = (last_irx - OSK_XSTART) / (osk_col_size * (osk_col_size / osk_charsize)) - 1;

		osk_coords[0] = last_irx;
		osk_coords[1] = last_iry;

		line = clamp(line, 0, osk_num_lines);
		col = clamp(col, 0, osk_num_col);

		if (nunchuk_connected && (sys_keys & WPAD_NUNCHUK_BUTTON_Z))
			osk_set = osk_shifted;
		else
			osk_set = osk_normal;

		osk_selected = osk_set[line * osk_num_col + col];

		if ((sys_keys & WPAD_BUTTON_B) && osk_selected && (Sys_FloatTime() >= osk_last_press_time + osk_repeat_delay || osk_selected != osk_last_selected))
		{
			Key_Event((osk_selected), true);
			Key_Event((osk_selected), false);
			osk_last_selected = osk_selected;
			osk_last_press_time = Sys_FloatTime();
		}
	}
	else
	{
		in_osk = 0;

		if ((sys_previous_keys & WPAD_BUTTON_LEFT) != (sys_keys & WPAD_BUTTON_LEFT))
		{
			// Send a press event.
			Key_Event(K_LEFTARROW, ((sys_keys & WPAD_BUTTON_LEFT) == WPAD_BUTTON_LEFT));
		}

		if ((sys_previous_keys & WPAD_BUTTON_RIGHT) != (sys_keys & WPAD_BUTTON_RIGHT))
		{
			// Send a press event.
			Key_Event(K_RIGHTARROW, ((sys_keys & WPAD_BUTTON_RIGHT) == WPAD_BUTTON_RIGHT));
		}

		if ((sys_previous_keys & WPAD_BUTTON_DOWN) != (sys_keys & WPAD_BUTTON_DOWN))
		{
			// Send a press event.
			Key_Event(K_DOWNARROW, ((sys_keys & WPAD_BUTTON_DOWN) == WPAD_BUTTON_DOWN));
		}

		if ((sys_previous_keys & WPAD_BUTTON_UP) != (sys_keys & WPAD_BUTTON_UP))
		{
			// Send a press event.
			Key_Event(K_UPARROW, ((sys_keys & WPAD_BUTTON_UP) == WPAD_BUTTON_UP));
		}

		if ((sys_previous_keys & WPAD_BUTTON_PLUS) != (sys_keys & WPAD_BUTTON_PLUS))
		{
			// Send a press event.
			Key_Event(K_ESCAPE, ((sys_keys & WPAD_BUTTON_PLUS) == WPAD_BUTTON_PLUS));
		}

		if ((sys_previous_keys & WPAD_BUTTON_A) != (sys_keys & WPAD_BUTTON_A))
		{
			// Send a press event.
			Key_Event(K_JOY8, ((sys_keys & WPAD_BUTTON_A) == WPAD_BUTTON_A));
		}

		if ((sys_previous_keys & WPAD_BUTTON_B) != (sys_keys & WPAD_BUTTON_B))
		{
			// Send a press event.
			Key_Event(K_JOY9, ((sys_keys & WPAD_BUTTON_B) == WPAD_BUTTON_B));
		}

		if ((sys_previous_keys & WPAD_BUTTON_MINUS) != (sys_keys & WPAD_BUTTON_MINUS))
		{
			// Send a press event.
			Key_Event(K_JOY10, ((sys_keys & WPAD_BUTTON_MINUS) == WPAD_BUTTON_MINUS));
		}

		if ((sys_previous_keys & WPAD_BUTTON_1) != (sys_keys & WPAD_BUTTON_1))
		{
			// Send a press event.
			Key_Event(K_JOY11, ((sys_keys & WPAD_BUTTON_1) == WPAD_BUTTON_1));
		}

		if ((sys_previous_keys & WPAD_BUTTON_2) != (sys_keys & WPAD_BUTTON_2))
		{
			// Send a press event.
			Key_Event(K_JOY12, ((sys_keys & WPAD_BUTTON_2) == WPAD_BUTTON_2));
		}

		if ((sys_previous_keys & WPAD_NUNCHUK_BUTTON_C) != (sys_keys & WPAD_NUNCHUK_BUTTON_C))
		{
			// Send a press event.
			Key_Event(K_JOY13, ((sys_keys & WPAD_NUNCHUK_BUTTON_C) == WPAD_NUNCHUK_BUTTON_C));
		}

		if ((sys_previous_keys & WPAD_NUNCHUK_BUTTON_Z) != (sys_keys & WPAD_NUNCHUK_BUTTON_Z))
		{
			// Send a press event.
			Key_Event(K_JOY14, ((sys_keys & WPAD_NUNCHUK_BUTTON_Z) == WPAD_NUNCHUK_BUTTON_Z));
		}
	}

	sys_previous_keys = sys_keys;
}

// Some things here rely upon IN_Move always being called after IN_Commands on the same frame
void IN_Move (usercmd_t *cmd)
{
	// Read the stick values.
// 	const s8 stick_x = PAD_StickX(0);
// 	const s8 stick_y = PAD_StickY(0);
// 	const s8 sub_stick_x = PAD_SubStickX(0);
// 	const s8 sub_stick_y = PAD_SubStickY(0);

	// TODO: new issue, if the wiimote gets resynced during game, sometimes we get invalid nunchuk data! Has it been fixed?
	const s8 nunchuk_stick_x = WPAD_StickX(WPAD_EXP_NUNCHUK, 0);
	const s8 nunchuk_stick_y = WPAD_StickY(WPAD_EXP_NUNCHUK, 0);
	// TODO: sensor bar position correct? aspect ratio correctly set? etc...
	int last_wiimote_ir_x = pointer.x;
	int last_wiimote_ir_y = pointer.y;
	int wiimote_ir_x = 0, wiimote_ir_y = 0;


	if (pointer.x < 1 || (unsigned int)pointer.x > pointer.vres[0] - 1)
		wiimote_ir_x = last_wiimote_ir_x;
	else
		wiimote_ir_x = pointer.x;
	if (pointer.y < 1 || (unsigned int)pointer.y > pointer.vres[1] - 1)
		wiimote_ir_y = last_wiimote_ir_y;
	else
		wiimote_ir_y = pointer.y;

	last_wiimote_ir_x = wiimote_ir_x;
	last_wiimote_ir_y = wiimote_ir_y;

	// Convert the inputs to floats in the range [-1, 1].
#if 0
	Con_Printf("pad: %4d, %4d %4d, %4d\n",
		pads[0].stickX,
		pads[0].stickY,
		pads[0].substickX,
		pads[0].substickY);
#endif

	if (in_osk || (cls.state == ca_connected && key_dest != key_game))
	{
		last_irx = wiimote_ir_x;
		last_iry = wiimote_ir_y;
		return;
	}

	// If the nunchuk is centered, read from the left gamecube pad stick
	float x1;
	float y1;
	float x2;
	float y2;
/*
	if (nunchuk_stick_x < 6 && nunchuk_stick_x > -6)
		x1 = clamp(stick_x / 90.0f, -1.0f, 1.0f);
	else*/
		x1 = clamp(((float)nunchuk_stick_x / 128.0f) * 1.5, -1.0f, 1.0f);
/*
	if (nunchuk_stick_y < 6 && nunchuk_stick_y > -6)
		y1 = clamp(stick_y / -90.0f, -1.0f, 1.0f);
	else*/
		y1 = clamp(((float)nunchuk_stick_y / (-128.0f)) * 1.5, -1.0f, 1.0f);

	// Now the gamecube C-stick has the priority
	static bool using_c_stick = false;
/*
	if ((sub_stick_x < 6 && sub_stick_x > -6) && wiimote_connected && (using_c_stick == false || (using_c_stick == true && last_irx != wiimote_ir_x)))
	{*/
		x2 = clamp((float)wiimote_ir_x / (pointer.vres[0] / 2.0f) - 1.0f, -1.0f, 1.0f);
		Cvar_SetValue("cl_crossx", scr_vrect.width / 2 * x2);
		using_c_stick = false;
// 	}
// 	else
// 	{
// 		x2 = clamp(sub_stick_x / 80.0f, -1.0f, 1.0f);
// 		Cvar_SetValue("cl_crossx", (in_mlook.state & 1) ? scr_vrect.width / 2 * x2 : 0);
// 		using_c_stick = true;
// 	}
// 
// 	if ((sub_stick_y < 6 && sub_stick_y > -6) && wiimote_connected && (using_c_stick == false || (using_c_stick == true && last_iry != wiimote_ir_y)))
// 	{
		y2 = clamp((float)wiimote_ir_y / (pointer.vres[1] / 2.0f) - 1.0f, -1.0f, 1.0f);
		Cvar_SetValue("cl_crossy", scr_vrect.height / 2 * y2);
		using_c_stick = false;
// 	}
// 	else
// 	{
// 		y2 = clamp(sub_stick_y / -80.0f, -1.0f, 1.0f);
// 		Cvar_SetValue("cl_crossy", (in_mlook.state & 1) ? scr_vrect.height / 2 * y2 : 0);
// 		using_c_stick = true;
// 	}

	last_irx = wiimote_ir_x;
	last_iry = wiimote_ir_y;

	// Apply the dead zone.
	const float dead_zone = 0.1f;
	apply_dead_zone(&x1, &y1, dead_zone);
	apply_dead_zone(&x2, &y2, dead_zone);

	// Don't let the pitch drift back to centre if mouse look is on or the right stick is being used.
	//if ((in_mlook.state & 1) || (fabsf(y2) >= dead_zone)) Disabled, always very convenient with a gamepad or wiimote
	{
		V_StopPitchDrift();
	}

	// Lock view?
	if (in_mlook.state & 1)
	{
		x2 = 0;
		y2 = 0;
	}

	float yaw_rate;
	float pitch_rate;
	// Turn using the substick.
	yaw_rate = x2;
	pitch_rate = y2;

// 	// Move using the main stick.
// 	cmd->sidemove += cl_sidespeed.value * x1;
// 	cmd->forwardmove -= cl_forwardspeed.value * y1; /* TODO: use cl_backspeed when going backwards? */
// 
// 	if (in_speed.state & 1)
// 	{
// 		if (cl_forwardspeed.value > 200)
// 		{
// 			cmd->forwardmove /= cl_movespeedkey.value;
// 			cmd->sidemove /= cl_movespeedkey.value;
// 		}
// 		else
// 		{
// 			cmd->forwardmove *= cl_movespeedkey.value;
// 			cmd->sidemove *= cl_movespeedkey.value; /* TODO: always seem to be at the max and I'm too sleepy now to figure out why */
// 		}
// 	}

	// TODO: Use yawspeed and pitchspeed

	// Adjust the yaw.
	const float turn_rate = sensitivity.value * 50.0f;
	if (in_speed.state & 1)
	{
		if (cl_forwardspeed.value > 200)
			cl.viewangles[YAW] -= turn_rate * yaw_rate * host_frametime / cl_anglespeedkey.value;
		else
			cl.viewangles[YAW] -= turn_rate * yaw_rate * host_frametime * cl_anglespeedkey.value;
	}
	else
		cl.viewangles[YAW] -= turn_rate * yaw_rate * host_frametime;

	// How fast to pitch?
	float pitch_offset;
	if (in_speed.state & 1)
	{
		if (cl_forwardspeed.value > 200)
			pitch_offset = turn_rate * pitch_rate * host_frametime / cl_anglespeedkey.value;
		else
			pitch_offset = turn_rate * pitch_rate * host_frametime * cl_anglespeedkey.value;
	}
	else
		pitch_offset = turn_rate * pitch_rate * host_frametime;

	// Do the pitch.
	const bool	invert_pitch = m_pitch.value < 0;
	if (invert_pitch)
	{
		cl.viewangles[PITCH] -= pitch_offset;
	}
	else
	{
		cl.viewangles[PITCH] += pitch_offset;
	}

	// Don't look too far up or down.
	if (cl.viewangles[PITCH] > 80.0f)
	{
		cl.viewangles[PITCH] = 80.0f;
	}
	else if (cl.viewangles[PITCH] < -70.0f)
	{
		cl.viewangles[PITCH] = -70.0f;
	}

	if (wiimote_connected && !using_c_stick)
	{
		in_pitchangle = orientation.pitch;
		in_yawangle = orientation.yaw;
		in_rollangle = orientation.roll;
	}
	else
	{
		in_pitchangle = .0f;
		in_yawangle = .0f;
		in_rollangle = .0f;
	}

	cmd->sidemove += m_side.value * (expansion.nunchuk.js.pos.x - expansion.nunchuk.js.center.x) * sensitivity.value;
	cmd->forwardmove -= m_forward.value * (expansion.nunchuk.js.center.y - expansion.nunchuk.js.pos.y) * sensitivity.value;
}
